/**
 * PaymentDisplay for ETH-based crypto payments.
 *
 * @author Patrick Müller (www.pcsg.de)
 */
define('package/quiqqer/payment-crypto-eth/bin/js/frontend/controls/PaymentDisplay', [

    'package/quiqqer/payment-crypto/bin/js/frontend/controls/PaymentDisplay',
    'package/quiqqer/authmetamask/bin/Web3',

    'Locale',
    'Ajax',

    'css!package/quiqqer/payment-crypto-eth/bin/js/frontend/controls/PaymentDisplay.css'

], function(ParentControl, Web3, QUILocale, QUIAjax) {
    'use strict';

    const pkg = 'quiqqer/payment-crypto-eth';

    return new Class({

        Extends: ParentControl,
        Type: 'package/quiqqer/payment-crypto-eth/bin/js/frontend/controls/PaymentDisplay',

        Binds: [
            '$onPayBtnClick',
            '$createTransactionNativeCurrency',
            '$createTransactionNonNativeToken',
            '$isTestMode'
        ],

        initialize: function(options) {
            this.parent(options);

            this.$contractAddress = null;
            this.$receiverAddress = null;
            this.$buyerAddress = null;
            this.$price = null;
            this.$web3 = null;
            this.$testMode = false;

            this.$prodNetworkId = 1;
            this.$testNetworkId = 5;
            this.$nativeCurrencySymbol = 'ETH';
        },

        $onPayBtnClick: function() {
            this.$contractAddress = this.getAttribute('contractaddress');
            this.$receiverAddress = this.getAttribute('receiveraddress');
            this.$price = this.getAttribute('price');

            let targetNetworkId = this.$prodNetworkId;

            this.$disablePayBtn();

            this.$isTestMode().then((isTestMode) => {
                this.$testMode = isTestMode;

                if (this.$testMode) {
                    targetNetworkId = this.$testNetworkId;
                }

                return Web3.getWeb3();
            }).then((web3) => {
                this.$web3 = web3;

                const buyerAddress = this.$web3.currentProvider.selectedAddress;

                if (typeof buyerAddress === 'undefined' || buyerAddress === '') {
                    this.$showErrorMsg(QUILocale.get(pkg, 'controls.PaymentDisplay.error.general'));
                    this.$onPaymentError();

                    return;
                }

                web3.eth.net.getId().then((network) => {
                    if (network === targetNetworkId) {
                        return;
                    }

                    return web3.currentProvider.request({
                        method: 'wallet_switchEthereumChain',
                        params: [
                            {
                                chainId: this.$web3.utils.toHex(targetNetworkId)
                            }
                        ]
                    }).catch((switchError) => {

                        // missing test network, add it
                        if (this.$testMode && switchError.code === 4902) {
                            return window.ethereum.request({
                                method: 'wallet_addEthereumChain',
                                params: [
                                    {
                                        chainId: this.$web3.utils.toHex(targetNetworkId),
                                        chainName: 'Goerli Test Network',
                                        nativeCurrency: {
                                            name: 'GoerliETH',
                                            symbol: 'GoerliETH',
                                            decimals: 18
                                        },
                                        rpcUrls: ['https://rpc.ankr.com/eth_goerli'],
                                        blockExplorerUrls: ['https://goerli.etherscan.io']
                                    }
                                ]
                            }).then(() => {
                                return web3.currentProvider.request({
                                    method: 'wallet_switchEthereumChain',
                                    params: [
                                        {
                                            chainId: this.$web3.utils.toHex(targetNetworkId)
                                        }
                                    ]
                                });
                            });
                        }

                        // This error code indicates that the chain has not been added to MetaMask.
                        if (switchError.code === 4902) {
                            this.$showErrorMsg(QUILocale.get(pkg, 'controls.PaymentDisplay.error.wrong_network'));
                            this.$onPaymentError();
                        }

                        return false;
                    });
                }).then(() => {
                    return web3.eth.net.getId();
                }).then((network) => {
                    if (network !== targetNetworkId) {
                        this.$showErrorMsg(QUILocale.get(pkg, 'controls.PaymentDisplay.error.wrong_network'));
                        this.$onPaymentError();

                        return;
                    }

                    this.$buyerAddress = buyerAddress;

                    switch (this.getAttribute('symbol')) {
                        case this.$nativeCurrencySymbol:
                            this.$createTransactionNativeCurrency();
                            break;

                        default:
                            this.$createTransactionNonNativeToken();
                    }
                });
            }).catch((e) => {
                console.error(e);

                this.$showErrorMsg(QUILocale.get(pkg, 'controls.PaymentDisplay.error.general'));
                this.$onPaymentError();
            });
        },

        /**
         * Create transaction with the "native" currency (ETH)
         *
         * @return {void}
         */
        $createTransactionNativeCurrency: function() {
            const transactionObject = {
                to: this.$receiverAddress,
                from: this.$buyerAddress,
                value: this.$web3.utils.toWei(this.$price.toString(), 'ether')
            };

            this.$web3.eth.estimateGas(transactionObject).then(estimatedGas => {
                transactionObject.gas = estimatedGas;

                return this.$web3.eth.sendTransaction(transactionObject);
            }).then((result) => {
                return this.$onTransaction(result.transactionHash);
            }).catch((err) => {

                let errorMsg;

                switch (err.code) {
                    case 4001:
                        errorMsg = QUILocale.get(pkg, 'controls.PaymentDisplay.error.transaction_cancelled');
                        break;

                    default:
                        errorMsg = QUILocale.get(pkg, 'controls.PaymentDisplay.error.general');
                }

                this.$showErrorMsg(errorMsg);
                this.$onPaymentError();
            });
        },

        /**
         * Create transaction with the non-"native" token (coins)
         *
         * @return {void}
         */
        $createTransactionNonNativeToken: function() {
            let abi = this.getAttribute('abi');

            if (!abi) {
                this.$showErrorMsg(QUILocale.get(pkg, 'controls.PaymentDisplay.error.general'));
                this.$onPaymentError();

                return;
            }

            abi = JSON.decode(abi);

            const Contract = new this.$web3.eth.Contract(
                abi,
                this.$contractAddress,
                {
                    from: this.$buyerAddress
                }
            );

            const amount = this.$web3.utils.toWei(this.$price.toString(), 'ether');

            Contract.methods.transfer(
                this.$receiverAddress,
                amount
            ).estimateGas({
                from: this.$buyerAddress
            }).then((estimatedGas) => {
                return Contract.methods.transfer(
                    this.$receiverAddress,
                    amount
                ).send({
                    to: this.$receiverAddress,
                    from: this.$buyerAddress,
                    gas: estimatedGas
                });
            }).then((result) => {
                return this.$onTransaction(result.transactionHash);
            }).catch((err) => {
                console.error(err);

                if (typeof err.code !== 'undefined') {
                    let errorMsg;

                    switch (err.code) {
                        case 4001:
                            errorMsg = QUILocale.get(pkg, 'controls.PaymentDisplay.error.transaction_cancelled');
                            break;

                        default:
                            errorMsg = QUILocale.get(pkg, 'controls.PaymentDisplay.error.general');
                    }

                    this.$showErrorMsg(errorMsg);
                    this.$onPaymentError();
                }
            });
        },

        /**
         * @return {Promise<Boolean>}
         */
        $isTestMode: function() {
            return new Promise((resolve, reject) => {
                QUIAjax.get('package_quiqqer_payment-crypto-eth_ajax_frontend_isTestMode', resolve, {
                    'package': pkg,
                    onError: reject,
                    showError: false
                });
            });
        }
    });
});