<?php

namespace QUI\ERP\Payments\Crypto\ETH;

use Exception;
use kornrunner\Keccak;
use stdClass;

use function array_map;
use function bin2hex;
use function implode;
use function is_string;
use function pow;
use function preg_match;
use function strlen;
use function substr;

/**
 * Ethereum ABI representation
 */
class Abi
{
    /**
     * Generates an event ID based on the given name and types.
     *
     * @param string $name The name of the event.
     * @param array $types An array of types associated with the event.
     *
     * @return string The generated event ID.
     * @throws Exception
     */
    public static function eventID(string $name, array $types): string
    {
        $types = implode(',', array_map('self::elementaryName', $types));
        $sig = $name . '(' . $types . ')';

        return Keccak::hash($sig, 256);
    }

    /**
     * Retrieves a portion of the event ID based on the given name and types.
     *
     * @param string $name The name to be used in generating the event ID.
     * @param array $types An array of types used in generating the event ID.
     *
     * @return string The portion of the event ID generated
     * @throws Exception
     */
    public static function methodID(string $name, array $types): string
    {
        return substr(self::eventID($name, $types), 0, 8);
    }

    /**
     * Decodes raw data using the given types and data inputs.
     *
     * @param array $types The array of data type strings to be decoded.
     * @param mixed $data The input data to be decoded.
     * @param mixed $dataBin The binary data to be decoded.
     *
     * @return array The decoded values as an array.
     * @throws Exception
     */
    public static function rawDecode(array $types, mixed $data, mixed $dataBin): array
    {
        $ret = [];
        $offset = 0;

        foreach ($types as $t) {
            $type = self::elementaryName($t);
            $parsed = self::parseType($type);
            $decoded = self::decodeSingle($parsed, $data, $dataBin, $offset);
            $offset += $parsed->memoryUsage;
            $ret[] = $decoded;
        }

        return $ret;
    }

    /**
     * Determines the elementary name for a given type.
     *
     * @param string $name The input type name.
     *
     * @return string The corresponding elementary name.
     */
    public static function elementaryName(string $name): string
    {
        if (str_starts_with($name, 'int[')) {
            return 'int256' . substr($name, 3);
        }

        if ($name === 'int') {
            return 'int256';
        }

        if (str_starts_with($name, 'uint[')) {
            return 'uint256' . substr($name, 4);
        }

        if ($name === 'uint') {
            return 'uint256';
        }

        if (str_starts_with($name, 'fixed[')) {
            return 'fixed128x128' . substr($name, 5);
        }

        if ($name === 'fixed') {
            return 'fixed128x128';
        }

        if (str_starts_with($name, 'ufixed[')) {
            return 'ufixed128x128' . substr($name, 6);
        }

        if ($name === 'ufixed') {
            return 'ufixed128x128';
        }

        return $name;
    }

    /**
     * Parses a string and extracts the numeric value at the end.
     *
     * @param string $type The input string to be parsed.
     *
     * @return int The numeric value extracted from the input string.
     */
    public static function parseTypeN(string $type): int
    {
        preg_match('/^\D+(\d+)$/', $type, $matches);
        return (int)$matches[1];
    }

    /**
     * Parses a string type in the format NxM and returns an array containing the values N and M.
     *
     * @param string $type The string type in the format NxM.
     * @return array The parsed N and M values as an array.
     */
    public static function parseTypeNxM(string $type): array
    {
        preg_match('/^\D+(\d+)x(\d+)$/', $type, $matches);
        return [(int)$matches[1], (int)$matches[2]];
    }

    /**
     * Decodes a single value based on the parsed type.
     *
     * @param mixed $parsedType The parsed type.
     * @param mixed $data The data to decode.
     * @param mixed $dataBin The binary data.
     * @param int $offset The offset.
     * @return mixed The decoded value.
     * @throws Exception If the type is unsupported or invalid.
     */
    public static function decodeSingle(mixed $parsedType, mixed $data, mixed $dataBin, int $offset): mixed
    {
        if (is_string($parsedType)) {
            $parsedType = self::parseType($parsedType);
        }

        switch ($parsedType->name) {
            case 'address':
            case 'string':
                return self::decodeSingle(
                    $parsedType->rawType,
                    $data,
                    $dataBin,
                    $offset
                );
            case 'bool':
                return (string)self::decodeSingle($parsedType->rawType, $data, $dataBin, $offset) === '1';
            case 'bytes':
                $offset = (int)self::decodeSingle('uint256', $data, $dataBin, $offset);
                $size = (int)self::decodeSingle('uint256', $data, $dataBin, $offset);
                return bin2hex(substr($dataBin, $offset + 32, ($offset + 32 + $size) - ($offset + 32)));
        }

        if ($parsedType->isArray) {
            $ret = [];
            $size = $parsedType->size;

            if ($parsedType->size === 'dynamic') {
                $offset = (int)self::decodeSingle('uint256', $data, $dataBin, $offset);
                $size = self::decodeSingle('uint256', $data, $dataBin, $offset);
                $offset = $offset + 32;
            }
            for ($i = 0; $i < $size; $i++) {
                $decoded = self::decodeSingle($parsedType->subArray, $data, $dataBin, $offset);
                $ret[] = $decoded;
                $offset += $parsedType->subArray->memoryUsage;
            }
            return $ret;
        }

        if (str_starts_with($parsedType->name, 'bytes')) {
            return bin2hex(substr($dataBin, $offset, ($offset + $parsedType->size) - $offset));
        }

        if (str_starts_with($parsedType->name, 'uint') || str_starts_with($parsedType->name, 'int')) {
            $num = bin2hex(substr($dataBin, $offset, $offset + 32 - $offset));

            if (strlen(bin2hex($num)) / 2 > $parsedType->size) {
                throw new Exception(
                    'Decoded int exceeds width: ' . $parsedType->size . ' vs ' . strlen(
                        bin2hex($num)
                    ) / 2
                );
            }

            return $num;
        }

        if (str_starts_with($parsedType->name, 'ufixed') || str_starts_with($parsedType->name, 'fixed')) {
            $size = pow(2, $parsedType->size[1]);
            $num = self::decodeSingle(
                str_starts_with($parsedType->name, 'ufixed') ? 'uint256' : 'int256',
                $data,
                $dataBin,
                $offset
            );
            if (!($num % $size === 0)) {
                throw new Exception('Decimals not supported yet');
            }
            return $num / $size;
        }

        throw new Exception('Unsupported or invalid type: ' . $parsedType->name);
    }

    /**
     * Parses a string and extracts the numeric value at the end.
     *
     * @param string $type The input string to be parsed.
     *
     * @return stdClass The numeric value extracted from the input string.
     * @throws Exception
     */
    public static function parseType(string $type): stdClass
    {
        $rawType = null;

        switch ($type) {
            case 'address':
                $rawType = 'uint160';
                break;
            case 'bool':
                $rawType = 'uint8';
                break;
            case 'string':
                $rawType = 'bytes';
                break;
        }

        $ret = new stdClass();
        $ret->isArray = false;
        $ret->rawType = $rawType;
        $ret->name = $type;
        $ret->memoryUsage = 32;

        if (
            str_starts_with($type, 'bytes') && $type !== 'bytes'
            || str_starts_with($type, 'uint')
            || str_starts_with($type, 'int')
        ) {
            $ret->size = self::parseTypeN($type);
        } else {
            if (
                str_starts_with($type, 'ufixed')
                || str_starts_with($type, 'fixed')
            ) {
                $ret->size = self::parseTypeNxM($type);
            }
        }

        if (
            str_starts_with($type, 'bytes')
            && $type !== 'bytes'
            && ($ret->size < 1 || $ret->size > 32)
        ) {
            throw new Exception('Invalid bytes<N> width: ' . $ret->size);
        }

        if (
            (str_starts_with($type, 'uint') || str_starts_with($type, 'int')) &&
            ($ret->size % 8 || $ret->size < 8 || $ret->size > 256)
        ) {
            throw new Exception('Invalid int/uint<N> width: ' . $ret->size);
        }

        return $ret;
    }
}
