<?php

namespace QUI\ERP\Payments\Crypto\ETH;

use Exception;

use function array_map;
use function array_reduce;
use function bin2hex;
use function file_exists;
use function file_get_contents;
use function hex2bin;
use function is_array;
use function str_repeat;
use function str_replace;
use function substr;
use function trim;

/**
 * Class Decoder
 */
class Decoder
{
    protected array $abi = [];

    /**
     * Constructs a new instance of the class.
     *
     * @param string|array $abiFile The path to the ABI file or the ABI array
     * @throws Exception If the ABI file does not exist.
     */
    public function __construct(string|array $abiFile)
    {
        if (is_array($abiFile)) {
            $this->abi = $abiFile;
            return;
        }

        if (!file_exists($abiFile)) {
            throw new Exception('Abi file not exists');
        }

        $this->abi = json_decode(file_get_contents($abiFile), true);
    }

    /**
     * Decodes the given data using the ABI file.
     *
     * @param string $data The data to decode.
     * @return array An array containing the decoded data.
     * @throws Exception
     */
    public function decodeData(string $data): array
    {
        $data = trim($data);
        $dataBuf = hex2bin(str_replace('0x', '', $data));
        $methodId = bin2hex(substr($dataBuf, 0, 4));

        $inputsBufBin = substr($dataBuf, 4);
        $inputsBuf = bin2hex($inputsBufBin);

        return array_reduce($this->abi, function ($acc, $obj) use ($methodId, $inputsBuf, $inputsBufBin) {
            if ($obj['type'] === 'constructor') {
                return $acc;
            }

            $name = $obj['name'] ?? 'null'; // The Buffer gets 'null' as string
            $types = isset($obj['inputs']) ? array_map(function ($input) {
                return $input['type'];
            }, $obj['inputs']) : [];

            $hash = Abi::methodID($name, $types);

            if ($hash === $methodId) {
                // https://github.com/miguelmota/ethereum-input-data-decoder/issues/8
                if ($methodId === 'a9059cbb') {
                    $inputsBuf = str_repeat('00', 12) .
                        bin2hex(substr($inputsBufBin, 12, 32 - 12)) .
                        bin2hex(substr($inputsBufBin, 32));
                }

                $inputs = Abi::rawDecode($types, $inputsBuf, $inputsBufBin);

                return [
                    'name' => $name,
                    'types' => $types,
                    'inputs' => $inputs
                ];
            }

            return $acc;
        });
    }
}
